from re import T
from .rea import pipeline, pipelines, stream, scopeCache

def parseJsons(aContent: str) -> list:
    ret = list()
    tt = aContent.split("}{")
    len_tt = len(tt)
    for i in range(len_tt):
        str = tt[i]
        if i > 0:
            str = "{" + str
        if i < len_tt - 1:
            str = str + "}"
        ret.append(str)
    return ret

def connectRemote(aLocal: str, aRemote: str, aWriteRemote, aClient: bool, aRemoteLocal: str):
    pipelines(aRemote)
    pipelines(aLocal).updateOutsideRanges({aRemote})
    pipelines().find("receiveFrom" + ("Server" if aClient else "Client")).next(aRemote + "_receiveRemote", (aLocal if aRemoteLocal == "" else aRemoteLocal) + ";any")
    pipelines().add(aWriteRemote, {"name": aRemote + "_sendRemote"})

class pipelineRemote(pipeline):
    def __init__(self, aRemoteName: str, aLocalName: str):
        super().__init__(aRemoteName)
        self.__m_localName = aLocalName
        
        def receiveRemote(aInput: stream):
            dt = aInput.data()
            cmd = dt.get("cmd", "")
            if cmd == "execute":
                scp = aInput.scope()
                scp_arr = dt.get("scope", [])
                len_scp = len(scp_arr)
                for i in range(0, len_scp - 1, 2):
                    scp.cache(scp_arr[i], scp_arr[i + 1])
                self.executeFromRemote(dt.get("name", ""), dt.get("data", None), dt.get("tag", ""), scp, dt.get("sync", dict({})), dt.get("needFuture", False), dt.get("from", ""))
            elif cmd == "remove":
                self.removeFromRemote(dt.get("name", ""))
            
        pipelines().add(receiveRemote, {"name": self.name() + "_receiveRemote"})
    
    def executeFromRemote(self, aName: str, aData, aTag: str, aScope: scopeCache, aSync: dict, aNeedFuture: bool, aFlag: str):
        pipelines(self.__m_localName).execute(aName, stream(aData, aTag, aScope), aSync, aNeedFuture, aFlag)

    def removeFromRemote(self, aName: str):
        pipelines(self.__m_localName).remove(aName, False)
    
    def remove(self, aName: str, aOutside: bool = False):
        pipelines().run(self.name() + "_sendRemote", {"cmd": "remove", "name": aName, "remote": "any"})
    
    def execute(self, aName: str, aStream: stream, aSync: dict = {}, aFutureNeed: bool = False, aFrom: str = ""):
        if aStream.scope().data("remote") is not None:
            lst = aStream.scope().toList()
            scp = []
            len_lst = len(lst)
            for i in range(0, len_lst - 1, 2):
                tp = type(lst[i + 1])
                if tp is str or tp is bool or tp is float or tp is int or tp is dict or tp is list:
                    scp.append(lst[i])
                    scp.append(lst[i + 1])
            pipelines().run(self.name() + "_sendRemote", {"cmd": "execute", 
                                                          "data": aStream.data(),
                                                            "name": aName,
                                                            "remote": self.name(),
                                                            "tag": aStream.tag(),
                                                            "scope": scp,
                                                            "sync": aSync,
                                                            "needFuture": aFutureNeed,
                                                            "from": self.name() if aFrom == "" else aFrom},
                                                            aStream.tag(),
                                                            aStream.scope())
                                                          




from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtNetwork import QHostAddress, QTcpServer, QTcpSocket
from .rea import pipelines, scopeCache
from .reaRemote import parseJsons
from time import sleep
import json

class normalServer(QObject):
    def __init__(self, aConfig: dict = {}):
        super().__init__(None)
        self.__m_socket = QTcpServer()
        self.__m_clients = set()
        self.__m_pipeline = aConfig.get("pipeline", "py")

        self.__m_socket.newConnection.connect(self.newConnect)
        self.__m_socket.listen(QHostAddress.LocalHost, 8081)

        pipelines(self.__m_pipeline).add(lambda aInput:
            aInput.out()
        , {"name": "receiveFromClient", "type": "Partial"})

        pipelines(self.__m_pipeline).add(lambda aInput:
            aInput.out()
        , {"name": "clientStatusChanged"})
        #self.connected = pyqtSignal()
    
    def __del__(self):
        self.__m_clients.clear()

    def disConnected(self):
        client = QObject.sender(self)
        pipelines(self.__m_pipeline).run("clientStatusChanged", False, "", scopeCache({"socket": client}))
        if client in self.__m_clients:
            self.__m_clients.remove(client)
            client.readyRead.disconnect(self.readMessage)
            client.disconnected.disconnect(self.disConnected)

    def newConnect(self):
        client = self.__m_socket.nextPendingConnection()
        client.readyRead.connect(self.readMessage)
        client.disconnected.connect(self.disConnected)
        pipelines(self.__m_pipeline).run("clientStatusChanged", True, "", scopeCache({"socket": client}))
        self.__m_clients.add(client)

    def readMessage(self):
        client = QObject.sender(self)
        while client.bytesAvailable():
            qba = client.readAll()
            strs = parseJsons(str(qba.data(), "utf-8"))
            for i in strs:
                req = json.loads(i)
                pipelines(self.__m_pipeline).run("receiveFromClient", req, req["remote"], scopeCache({"socket": client}))

    def writeSocket(self, aSocket: QTcpSocket, aData: dict):
        if aSocket in self.__m_clients:
            aSocket.write(bytearray(json.dumps(aData).encode("utf-8")))
            aSocket.flush()
            while aSocket.bytesToWrite() > 0:
                sleep(0.005)